import argutils
import argutils.read
import argutils.export
import swga.utils as utils
import activate
import count
import export
import filter
import find_sets
import score
import summary


_commands = [
    'activate',
    'count',
    'export',
    'filter',
    'find_sets',
    'score',
    'summary',
]


def create_config_file():
    cfg_file_str = ""
    for cmd in _commands:
        spec = utils.specfile(cmd)
        opts = argutils.read.from_yaml(spec)
        if opts:
            if opts['_meta'].get("_exclude"):
                continue
            cfg_file_str += argutils.export.to_config(cmd, opts) + "\n"
    return cfg_file_str


Activate = activate.Activate
Count = count.Count
Export = export.Export
Filter = filter.Filter
FindSets = find_sets.FindSets
Score = score.Score
Summary = summary.Summary



"""stats.py

Functions used to score primer sets.

"""

from math import sqrt


def mean(values):
    """Calculate the arithmetic mean."""
    return float(sum(values)) / len(values) if len(values) > 0 else float('nan')


def stdev(values):
    """Calculate the corrected sample standard deviation."""
    n = float(len(values))
    mu = sum(values) / n
    return sqrt(sum((x - mu)**2 for x in values) / (n - 1))


def seq_diff(seq):
    """Calculate the sequential differences of a set of numbers.

    :param seq: a set of numbers, will be sorted before sequential differences
    """
    seq.sort()
    diffs = []
    for i in xrange(len(seq) - 1):
        diff = seq[i + 1] - seq[i]
        assert diff >= 0
        diffs.append(diff)
    return diffs


def gini(distances):
    """Calculate an approximation of the Gini coefficient for a set of values.

    The Gini coefficient represents the evenness of a set of values, and is in
    the range [0,1], where 0 is perfectly even and 1 is perfectly uneven.

    In practice, this means that sets with low Gini coefficients have more even
    binding frequencies and are more likely to not favor highly-repeated regions
    or other aberrations in the target genome.

    :param distances: a vector of primer binding distances (from each other)
    """
    distances.sort()
    height = area = 0
    for value in distances:
        height += value
        area += height - value / 2.
    fair_area = height * len(distances) / 2
    return (fair_area - area) / fair_area


def lorenz(distances):
    """Calculate the Lorenz curve for a set of values.

    The Lorenz curve is the cumulative total value of a set of values ordered
    by smallest to largest. The more this deflects from the line of perfect
    equality, the higher the relative inequality of the values (the Gini
    coefficient. These numbers are provided to plot the Lorenz curve for
    visualization purposes.

    For more information, see https://en.wikipedia.org/wiki/Lorenz_curve.

    :param distances: a vector of primer binding distances (from each other)
    """
    distances.sort()
    height = 0
    lz = []
    for dist in distances:
        height += dist
        lz.append(height)

    # Normalize output
    _max = float(max(lz))
    lz = [l / _max for l in lz]

    return lz



# -*- coding: utf-8 -*-
"""locate.py

This module contains functions for locating primers in genomes (i.e. finding
small substrings in larger strings).

"""
import json
from pyfaidx import Fasta


def revcomp(s):
    r = {
        'A': 'T', 'T': 'A',
        'G': 'C', 'C': 'G'
    }
    rc = [r[i] for i in s][::-1]
    return "".join(rc)


def linearize_binding_sites(primers, chr_ends):
    '''
    Modifies the primer binding site locations as if they were positions on a
    linear genome composed of all the chromosomes concatenated together. This
    allows us to compute distances between primer binding sites correctly.
    '''
    new_locs = []
    for primer in primers:
        for rec, locs in primer.locations.iteritems():
            chr_start, chr_end = chr_ends[rec]
            new_locs += [l + chr_start for l in locs] + [chr_start, chr_end]
    new_locs = list(set(new_locs))
    if new_locs == []:
        raise ValueError("Binding sites for primers not found!")
    return list(set(new_locs))


def binding_sites(kmer, genome_fp):
    genome = Fasta(genome_fp)
    locations = {}
    kmer = str(kmer)
    for record in genome.keys():
        seq = str(genome[record])
        locations[record] = substr_indices(kmer, seq)
        # append reversed primer locations as well
        locations[record] += substr_indices(revcomp(kmer), seq)
    if locations == {}:
        raise ValueError(
            "No locations for {} found in fg genome!".format(kmer))
    return locations


def _primer_bind_sites(primer, genome_fp):
    primer.locations = json.dumps(binding_sites(primer.seq, genome_fp))
    return primer


def chromosome_ends(genome_fp):
    '''
    Returns the locations of the starts/ends of each chromosome (record) in a
    genome where all the chromosomes are concatenated (so i.e. the 2nd genome
    start site is len(1st genome), and all indices are 0-based).
    '''
    genome = Fasta(genome_fp)
    len_so_far = 0
    chr_ends = {}
    for record in genome.keys():
        chromosome = genome[record]
        chr_len = len(chromosome)
        chr_ends[record] = [len_so_far, chr_len + len_so_far - 1]
        len_so_far += chr_len
    return chr_ends


def substr_indices(substring, string):
    '''
    Very fast way of finding overlapping substring locations in a
    (potentially large) string.
    '''
    locations = []
    start = 0
    # Assumes string is all upper-case
    substring = substring.upper()
    while True:
        start = string.find(substring, start) + 1
        if start > 0:
            locations.append(start-1)
        else:
            return locations



import argparse
import os
from swga import (
    error,
    DEFAULT_DB_FNAME,
    DEFAULT_CFG_FNAME
)
import swga.database as database
from swga.commands import (
    init,
    Summary,
    Count,
    Filter,
    FindSets,
    Score,
    Activate,
    Export
)

usage = """Usage: swga <command> [options]

Utility commands:
  init:             initializes a directory with a pre-filled parameters file
  summary:          get a summary of the primers and sets found so far

Pipeline commands:
  count:            find kmer counts in foreground and background genomes
  filter:           filter kmers according to various criteria
  find_sets:        find and score compatible sets of primers in graph
  export:           export sets and primers

Other commands:
  activate:         activate a list of primers, calculating Tm and locations
  score:            score a set specified by a list of primers in a file
"""


def main():
    command_opts = {
        'init': init.main,
        'summary': Summary,
        'count': Count,
        'filter': Filter,
        'find_sets': FindSets,
        'score': Score,
        'activate': Activate,
        'export': Export
    }

    parser = argparse.ArgumentParser(
        usage=usage,
        formatter_class=argparse.RawDescriptionHelpFormatter,
        add_help=False)

    parser.add_argument(
        'command',
        type=str,
        choices=command_opts.keys())

    args, remaining = parser.parse_known_args()

    try:
        if args.command == 'init':
            init.main(remaining)
        else:
            db_name = os.path.abspath(DEFAULT_DB_FNAME)
            cfg_file = os.path.abspath(DEFAULT_CFG_FNAME)
            with database.connection(db_name):
                database.check_version()
                metadata = database.get_metadata(db_name)
                cmd = command_opts[args.command](args.command, cfg_file, metadata)
                cmd.parse_args(remaining)
                cmd.run()
    except KeyboardInterrupt:
        error("\n-- Stopped by user --", exception=False)



from swga import warn
from swga.primers import Primers
from swga.commands._command import Command


class Activate(Command):

    def run(self):
        primers = Primers(self.input)

        try:
            (primers
                .update_melt_temps()
                .update_locations(self.fg_genome_fp)
                .activate())
        except AttributeError as e:
            warn("Error updating database: '{}'".format(e.message))
            raise e



# -*- coding: utf-8 -*-
'''Functions for counting/reading raw k-mer sequences.'''
from __future__ import with_statement, division
import subprocess
import os
import re
import struct

from swga import message
#from swga.utils import dsk
import swga.utils


def parse_kmer_file(lines):
    '''Returns the first field of whitespace-delimited lines.

    :param lines: an iterable of strings (e.g. an open handle)
    '''
    seqs = [re.split(r'[ \t]+', line.strip('\n'))[0] for line in lines]
    return seqs


def count_kmers(k, genome_fp, cwd, threshold=1):
    '''Counts k-mers in the specified genome.

    :param k: the number of nucleotides in the k-mers
    :param genome_fp: the file path to the genome/fasta file
    :param cwd: the current working directory (to store intermediate cache)
    :param threshold: the minimum k-mer frequency
    '''

    assert isinstance(threshold, int)

    genome = genome_fp.split(os.sep).pop()
    out = '%s-%dmers' % (genome, k)
    outfile = os.path.join(cwd, out + '.solid_kmers_binary')

    if os.path.isfile(outfile):
        message("Binary kmer file already exists at %s; parsing..."
                % outfile)
    else:
        cmdarr = [swga.utils.dsk, genome_fp, str(k), '-o', out, '-t', str(threshold)]
        cmdstr = " ".join(cmdarr)
        message("In {cwd}:\n> {cmdstr}".format(**locals()))
        try:
            subprocess.check_call(cmdarr, cwd=cwd)
        except:
            if os.path.isfile(outfile):
                os.remove(outfile)
            raise

    primers = dict((kmer, freq)
                   for kmer, freq in _parse_kmer_binary(outfile)
                   if freq >= threshold)

    return primers


def max_sequential_nt(mer1, mer2):
    '''Returns the maximum number of nt that would bind between two kmers.

    :param kmer1: the first kmer
    :param kmer2: the second kmer (order unimportant)
    '''

    binding = { 'A': 'T', 'T': 'A',
                'C': 'G', 'G': 'C',
                '_':  False}

    # Swap variables if the second is longer than the first
    if len(mer2) > len(mer1):
        mer1, mer2 = mer2, mer1

    # save the len because it'll change when we do a ljust
    mer1_len = len(mer1)
    # reverse mer2,
    mer2 = mer2[::-1]
    # pad mer one to avoid errors
    mer1 = mer1.ljust(mer1_len + len(mer2), "_")

    max_bind = 0
    for offset in range(mer1_len):
        consecutive = 0
        for x in range(len(mer2)):
            if binding[mer1[offset+x]] == mer2[x]:
                consecutive += 1
                if consecutive > max_bind:
                    max_bind = consecutive
            else:
                consecutive = 0
    return max_bind


def _parse_kmer_binary(fp):
    # Adapted from `dsk/parse_results.py`
    with open(fp, 'rb') as f:
        try:
            kmer_nbits = struct.unpack('i', f.read(4))[0]
            k = struct.unpack('i', f.read(4))[0]
        except struct.error:
            if os.path.isfile(fp):
                os.remove(fp)
            raise
        try:
            while True:
                kmer_binary = struct.unpack('B' * (kmer_nbits // 8),
                                            f.read(kmer_nbits // 8))
                freq = struct.unpack('I', f.read(4))[0]
                kmer = ""
                for i in xrange(k):
                    kmer = "ACTG"[
                        (kmer_binary[i // 4] >> (2 * (i % 4))) % 4] + kmer
                yield kmer, freq
        except struct.error:
            pass




# -*- coding: utf-8 -*-
"""
database.py

Defines the models and fields of the primer database using Peewee ORM. Also
contains a litany of helper functions for adding and retrieving stored data.

"""
import json
import os
from contextlib import contextmanager
import click
from swga import (error, warn, meta, __version__)
from swga.utils import chunk_iterator
import swga.stats as stats
import swga.locate as locate
import melting
import semantic_version as semver
import peewee as pw
try:
    from playhouse.shortcuts import ManyToManyField
except ImportError:
    from playhouse.fields import ManyToManyField

# The primer database must be initialized before use
# ex: `db.init(db_fname)`
db = pw.SqliteDatabase(None)


class SwgaBase(pw.Model):

    '''Specifies what database the inheriting models will use.'''

    @classmethod
    def fieldnames(cls):
        '''Returns the fields defined in the model as a list of strings.'''
        return [name for name, attr in cls.__dict__.items()
                if isinstance(attr, pw.FieldDescriptor)]

    @classmethod
    def fields(cls):
        return dict((name, attr.field) for name, attr in cls.__dict__.items()
                    if isinstance(attr, pw.FieldDescriptor))

    def to_dict(self):
        return self.__dict__['_data']

    class Meta:
        database = db


class Primer(SwgaBase):

    '''
    The primers table contains the sequence and metadata for each primer. Once
    set composition is determined, the sets that each primer belongs to can be
    found by using the PrimerSet intermediate table.
    '''
    _id = pw.IntegerField(null=True)
    seq = pw.CharField(primary_key=True)
    fg_freq = pw.IntegerField(default=0)
    bg_freq = pw.IntegerField(default=0)
    ratio = pw.FloatField(default=0.0)
    tm = pw.FloatField(null=True)
    _locations = pw.TextField(null=True)
    gini = pw.FloatField(null=True)
    active = pw.BooleanField(default=False)

    @staticmethod
    def exported_fields():
        fields = [
            'seq',
            'fg_freq',
            'bg_freq',
            'ratio',
            'gini',
            'tm']
        return fields

    def __repr__(self):
        rep_str = "Primer {0}:{1} (fg_freq:{2}, bg_freq:{3}, ratio:{4})"
        return rep_str.format(
            self.id, self.seq, self.fg_freq, self.bg_freq, self.ratio)

    @property
    def locations(self):
        if self._locations:
            return json.loads(self._locations)
        else:
            error("No locations stored for " + str(self))

    def update_tm(self):
        self.tm = melting.temp(
            self.seq, dNTPs_c=1, Mg_c=0.8, Na_c=50, DNA_c=200)

    def _update_locations(self, genome_fp):
        self._locations = json.dumps(
            locate.binding_sites(self.seq, genome_fp))

    def _update_gini(self, genome_fp):
        chr_ends = locate.chromosome_ends(genome_fp)
        locs = locate.linearize_binding_sites([self], chr_ends)
        dists = stats.seq_diff(locs)
        self.gini = stats.gini(dists)


class Set(SwgaBase):

    '''
    The sets table contains each set's metadata. The actual primers that belong
    in the set are found using the PrimerSet intermediate table.
    '''
    _id = pw.PrimaryKeyField()
    _hash = pw.IntegerField(unique=True, null=True)
    primers = ManyToManyField(Primer, related_name='sets')
    score = pw.FloatField()
    set_size = pw.IntegerField(null=True)
    bg_dist_mean = pw.FloatField(null=True)
    fg_dist_std = pw.FloatField(null=True)
    fg_dist_gini = pw.FloatField(null=True)
    fg_max_dist = pw.IntegerField(null=True)
    fg_dist_mean = pw.FloatField(null=True)
    scoring_fn = pw.TextField(null=True)

    def __repr__(self):
        attr_str = "; ".join(
            "{}:{}".format(k, v) for k, v
            in self.__dict__['_data'].items())
        return "Set: " + attr_str

    @staticmethod
    def exported_fields():
        fields = [
            '_id',
            'score',
            'set_size',
            'bg_dist_mean',
            'fg_max_dist',
            'fg_dist_mean',
            'fg_dist_std',
            'fg_dist_gini',
            'scoring_fn',
            'primers'
        ]
        return fields


PrimerSet = Set.primers.get_through_model()


class _metadata(SwgaBase):

    '''Holds metadata about the swga workspace.'''
    db_name = pw.TextField()
    version = pw.TextField()
    fg_file = pw.TextField()
    bg_file = pw.TextField()
    ex_file = pw.TextField()
    fg_length = pw.IntegerField()
    bg_length = pw.IntegerField()

@contextmanager
def connection(db_name, create_if_missing=False):
    if db_name != ':memory:':
        if not os.path.isfile(db_name) and not create_if_missing:
            error(
                '{} does not appear to be a workspace (database not found). '
                'Run `swga init` to initialize this directory as a workspace.'
                .format(os.path.curdir))
    db.init(db_name)
    db.connect()
    yield db
    db.close()


def init_db(db_fname, create_if_missing=False):
    '''
    Initializes the database at the file path specified.
    If `create_if_missing` is True, it will create the database if it can't be
    found. Otherwise, it exits with an error (SystemExit).
    '''
    if db_fname is None:
        error("Database name cannot be `None'.")
    elif db_fname == ":memory:":
        warn("Creating in-memory database.")
    elif not os.path.isfile(db_fname) and not create_if_missing:
        # Exits here
        error(
            "Database not found at '%s': specify different filename or "
            "re-run `swga count`" % db_fname, exception=False
        )
    db.init(db_fname)
    return db


def create_tables(drop=True):
    '''
    Creates the tables in the database. If `drop`=True, attempts to safely drop
    the tables before creating them.
    '''
    tbl_list = [Primer, Set, PrimerSet, _metadata]
    if drop:
        db.drop_tables(tbl_list, safe=True)
    db.create_tables(tbl_list, safe=True)


def check_create_tables(primer_db, skip_check=False):
    if os.path.isfile(primer_db) and not skip_check:
        warn(
            "This directory was already initialized as a workspace. "
            "Continuing will reset any stored primers and sets you may have found."
        )
        click.confirm("Are you sure you want to proceed?", abort=True)
    create_tables()


def check_version(swga_version=__version__):
    try:
        meta = _metadata.get()
        db_ver = semver.Version(meta.version)
    except pw.OperationalError:
        db_ver = "<NA>"
    ver = semver.Version(swga_version)
    spec = semver.Spec('=={}.{}'.format(ver.major, ver.minor))
    if db_ver not in spec:
        error(
            "This workspace was created with an incompatible version of swga.\n"
            "  Workspace version: {}\n"
            "  swga version:      {}\n"
            "Please re-initialize the workspace with `swga init` or use a "
            "different version of swga."
            .format(db_ver, ver),
            exception=False,
            wrap=False
        )


def add_primers(primers, chunksize=199, add_revcomp=True):
    if add_revcomp:
        def mkrevcomp(p):
            p2 = dict(**p)
            p2['seq'] = locate.revcomp(p['seq'])
            return p2
        primers += [mkrevcomp(p) for p in primers]
    chunk_iterator(
        primers,
        fn=lambda c: Primer.insert_many(c).execute(),
        n=chunksize,
        label="Updating database: ")


@db.atomic()
def add_set(_id, primers, **kwargs):
    if not primers:
        error("Invalid primers for set")
    if isinstance(primers, pw.SelectQuery):
        nprimers = primers.count()
    else:
        nprimers = len(primers)
    if nprimers == 0:
        error("Cannot have an empty set")
    _hash = hash(frozenset([p.seq for p in primers]))
    if Set.select(pw.fn.Count(Set._id)).where(Set._hash == _hash).scalar() > 0:
        return None
    s = Set.create(_id=_id, _hash=_hash, **kwargs)
    try:
        s.primers.add(primers)
    # If this is a Primers object, return the actual primers in it
    except AttributeError:
        s.primers.add(primers.primers)
    return s


def get_primers_for_set(set_id):
    set = Set.get(Set._id == set_id)
    return [p.seq for p in set.primers]


def get_primers_for_ids(pids):
    return list(Primer.select().where(Primer._id << pids).execute())


def get_metadata(db_name):
    m = _metadata.get()
    return meta(m.db_name, m.fg_file, m.bg_file, m.ex_file, m.fg_length, m.bg_length)

def set_metadata(db_name, version, fg_file, bg_file, ex_file, fg_length, bg_length):
    """Sets the metadata for a workspace. Clears any previous info."""
    _metadata.delete().execute()
    _metadata.insert(
        db_name=db_name,
        version=version,
        fg_file=fg_file,
        bg_file=bg_file,
        ex_file=ex_file,
        fg_length=fg_length,
        bg_length=bg_length
    ).execute()




# -*- coding: utf-8 -*-
"""score.py

Functions for retrieving and scoring primer sets.

"""

import importlib

import locate
import stats

from swga import warn


def read_set_finder_line(line):
    """Read a line from set_finder and convert it to a tuple of values.

    :param line: a string in the format "primer_id1,primer_id2,... weight"
    :returns: (size, weight, [primer_id1, primer_id2, ...]).
    """
    primer_set, weight = line.strip('\n').split(' ')
    primer_set = [int(_) for _ in primer_set.split(',')]
    return (primer_set, float(weight))


def default_score_set(expression, primer_set, primer_locs, max_dist, bg_dist_mean):
    """Evaluate an expression using the provided values and a set of metrics.

    :returns: the score and the metrics used to calculate it
    """
    # Calculate various metrics
    binding_distances = stats.seq_diff(primer_locs)
    namespace = {
        'set_size': len(primer_set),
        'fg_dist_mean': stats.mean(binding_distances),
        'fg_dist_std': stats.stdev(binding_distances),
        'fg_dist_gini': stats.gini(binding_distances),
        'bg_dist_mean': bg_dist_mean,
        'fg_max_dist': max_dist,
        '__builtins__': None}
    permitted_var_str = ", ".join(
        [key for key in namespace.keys() if key is not "__builtins__"])
    score = None
    try:
        score = eval(expression, namespace, {'__builtins__': {}})
    except NameError as e:
        raise NameError(
            e.message +
            '. Permitted variables are %s. Refer to README or docs for help.'
            % permitted_var_str)
    del namespace['__builtins__']
    return score, namespace


def calculate_bg_dist_mean(primers, bg_length):
    """Calculate the mean distance between binding sites on the bg genome.

    :param bg_length: the total length of the background genome.
    """
    total_bg_freq = sum(p.bg_freq for p in primers)
    if total_bg_freq == 0:
        warn(
            "No primers appear in the background genome: "
            "bg_dist_mean set as infinite")
        bg_dist_mean = float('Inf')
    else:
        bg_dist_mean = float(
            bg_length) / sum(p.bg_freq for p in primers)
    return bg_dist_mean


def score_set(primers, max_fg_bind_dist, bg_dist_mean,
              chr_ends, score_fun, interactive=False):
    """Score a set using the provided `score_fun`.

    :param bg_dist_mean: the average distance between binding sites on the
    background genome.
    :param chr_ends: the start and ends of each record in the foreground genome.
    :param score_fun: the scoring function
    :param interactive: if True, don't abort early due to set not passing filter
    """
    binding_locations = locate.linearize_binding_sites(
        primers, chr_ends)
    max_dist = max(stats.seq_diff(binding_locations))

    # If it's not a user-supplied set and it's not passing the filter,
    # abort immediately
    if not interactive and max_dist > max_fg_bind_dist:
        return False, {}, max_dist

    set_score, variables = score_fun(
        primer_set=primers,
        primer_locs=binding_locations,
        max_dist=max_dist,
        bg_dist_mean=bg_dist_mean)

    return set_score, variables, max_dist



import click
import functools
from peewee import fn
from swga import error, message
import swga.utils as utils
import swga.score as score
import swga.database
from swga.primers import Primers
from swga.commands._command import Command
from swga.database import Set


class Score(Command):

    def run(self):
        self.chr_ends = swga.locate.chromosome_ends(self.fg_genome_fp)
        # Evaluate the scoring expression from a string and return it as a
        # callable function
        self.score_fun = functools.partial(
            swga.score.default_score_set,
            expression=self.score_expression)

        primers = Primers(self.input)
        if len(primers) == 0:
            error(
                "No primers specified exist in database, aborting.",
                exception=False)

        bg_dist_mean = score.calculate_bg_dist_mean(primers, self.bg_length)

        set_score, variables, _ = score.score_set(
            primers=primers,
            max_fg_bind_dist=0,
            bg_dist_mean=bg_dist_mean,
            chr_ends=self.chr_ends,
            score_fun=self.score_fun,
            interactive=True
        )

        do_add_set, set_id = self.user_add_set(set_score, variables)

        if do_add_set:
            s = swga.database.add_set(
                _id=set_id,
                primers=primers,
                score=set_score,
                scoring_fn=self.score_expression,
                **variables)
            set_added = s is not None

            if set_added:
                swga.message("Set {} added successfully.".format(set_id))
            else:
                swga.message("That primer set already exists.")

    def user_add_set(self, set_score, variables):
        """Output set statistics and prompt the user to add the set."""
        set_dict = dict(
            {'score': set_score,
             'scoring_fn': self.score_expression}.items() +
            variables.items())
        message("Set statistics:\n - " + "\n - ".join(
            utils.fmtkv(k, v) for k, v in set_dict.items()))
        if self.force or (not self.force and click.confirm("Add set to database?",
                                                           default=True)):
            # User-provided sets have negative numbers, so we find the
            # smallest and decrement by 1
            min_set_id = Set.select(fn.Min(Set._id)).scalar()
            # This is None if there are no other sets yet
            if min_set_id is None:
                min_set_id = 0
            set_id = min_set_id - 1
            add_set = True
        else:
            add_set = False
        return add_set, set_id



# -*- coding: utf-8 -*-
import os
from collections import defaultdict

from swga import (error, message)
import swga.database as database
import swga.kmers
from swga.primers import Primer
from swga.commands._command import Command
from swga.utils import mkdirp
from peewee import OperationalError

INF = float('inf')
output_dir = ".swga_tmp"


class Count(Command):

    def run(self):
        if self.input:
            kmers = swga.kmers.parse_kmer_file(self.input)
            self.count_specific_kmers(kmers)
        else:
            self.count_kmers()

    def count_specific_kmers(self, kmers):
        try:
            # Skip primers that already exist and warn users
            existing = [p.seq for p in Primer.select().where(Primer.seq << kmers)]
            for p in existing:
                message("{} already exists in db, skipping...".format(p))
            kmers = [p for p in kmers if p not in existing]
        except OperationalError:
            # If this fails due to an OperationalError, it probably means the
            # database tables haven't been created yet. 
            error(
                "It doesn't appear that the workspace has been initialized: "
                "run `swga init' first.")
        mkdirp(output_dir)

        # Group the kmers by length to avoid repeatedly counting kmers of the
        # same size
        kmers_by_length = defaultdict(list)
        for kmer in kmers:
            kmers_by_length[len(kmer)].append(kmer)

        for k, mers in kmers_by_length.items():
            fg = swga.kmers.count_kmers(k, self.fg_genome_fp, output_dir, 1)
            bg = swga.kmers.count_kmers(k, self.bg_genome_fp, output_dir, 1)
            primers = []
            for mer in mers:
                try:
                    primers.append(primer_dict(mer, fg, bg, 0, INF, INF))
                except KeyError:
                    message(
                        "{} does not exist in foreground genome, skipping..."
                        .format(mer))

            # Omitting any primers that were returned empty
            # primers = filter(lambda p: p == {}, primers)
            chunk_size = 199
            message(
                "Writing {n} {k}-mers into db in blocks of {cs}..."
                .format(n=len(primers), k=k, cs=chunk_size))
            database.add_primers(primers, chunk_size, add_revcomp=False)

    def count_kmers(self):
        mkdirp(output_dir)

        kmers = []
        for k in xrange(self.min_size, self.max_size + 1):
            fg = swga.kmers.count_kmers(k, self.fg_genome_fp, output_dir)
            bg = swga.kmers.count_kmers(k, self.bg_genome_fp, output_dir)

            if self.exclude_fp:
                assert os.path.isfile(self.exclude_fp)
                ex = swga.kmers.count_kmers(
                    k, self.exclude_fp, output_dir, self.exclude_threshold)
            else:
                ex = {}

            # Keep kmers found in foreground, merging bg binding values, and
            # excluding those found in the excluded fasta

            kmers = [
                primer_dict(seq, fg, bg, self.min_fg_bind, self.max_bg_bind,
                            self.max_dimer_bp)
                for seq in fg.viewkeys() if seq not in ex.viewkeys()
            ]

            kmers = filter(lambda x: x != {}, kmers)

            nkmers = len(kmers)

            chunk_size = 199
            message(
                "Writing {n} {k}-mers into db in blocks of {cs}..."
                .format(n=nkmers * 2, k=k, cs=chunk_size))
            database.add_primers(kmers, chunk_size, add_revcomp=True)

        message("Counted kmers in range %d-%d" % (self.min_size, self.max_size))


def primer_dict(seq, fg, bg, min_fg_bind, max_bg_bind, max_dimer_bp):
    fg_freq = fg[seq]
    bg_freq = bg.get(seq, 0)
    ratio = fg_freq / float(bg_freq) if bg_freq > 0 else float('inf')

    # Setting to -1 disables the background binding frequency check
    max_bg_bind = INF if max_bg_bind < 0 else max_bg_bind

    if ((fg_freq >= min_fg_bind) and
            (bg_freq <= max_bg_bind) and
            # Homodimer check
            (swga.kmers.max_sequential_nt(seq, seq) <= max_dimer_bp)):
        return {
            'seq': seq,
            'fg_freq': fg_freq,
            'bg_freq': bg_freq,
            'ratio': ratio}
    else:
        return {}



'''
Functions for handling lists of primers (filtering, updating, etc).
'''

import re
from functools import wraps
from swga import (error, warn, message)
from swga.utils import chunk_iterator
from swga.database import Primer
from peewee import SelectQuery


def _filter(fn):
    """Wrapper for filter methods.

    Updates the internal count of primers and ensures the statement
    executes after each filter method.
    """
    @wraps(fn)
    def func(self, *args, **kwargs):
        results = fn(self, *args, **kwargs)
        assert isinstance(results, SelectQuery)
        self.primers = results
        self._update_n()
        return self
    return func


def _update(fn):
    """Wrapper for update methods.

    Updates the database in chunks after each update method.
    """
    @wraps(fn)
    def func(self, *args, **kwargs):
        targets = fn(self, *args, **kwargs)
        show_progress = len(targets) > 300
        Primers._update_in_chunks(targets, show_progress=show_progress)
        return self
    return func


class Primers(object):

    """A list of primers, and self-modifying filter/update functions."""

    @staticmethod
    def select_active():
        active = Primer.select().where(Primer.active == True)
        if active.count() == 0:
            error(
                'No active primers found. Run `swga filter` or `swga activate` '
                'first.',
                exception=False)
        return Primers(active)

    def __init__(self, primers=None):
        """Create a new list of primers.

        :param primers: a list of Primer objects or a list of primer sequences.
        If None, selects all primers.
        """
        if primers is None:
            self._primers = Primer.select()
            self.n = self.primers.count()
        elif isinstance(primers, file):
            self._primers = read_primer_list(primers)
            self.n = len(self.primers)
        else:
            self._primers = Primer.select().where(Primer.seq << primers)
            self.n = self.primers.count()

    @property
    def primers(self):
        return self._primers

    @primers.setter
    def primers(self, value):
        self._primers = value

    def __len__(self):
        return self.n

    def __getitem__(self, key):
        return self.primers[key]

    def __iter__(self):
        query = (
            Primer.select()
            .where(Primer.seq << self.primers)
            .order_by(Primer._id)
        )
        return query.iterator()

    @_filter
    def filter_min_fg_rate(self, min_bind):
        """Remove primers that bind less than `min_bind` to the foreground."""
        results = Primer.select().where(
            (Primer.seq << self.primers) &
            (Primer.fg_freq >= min_bind))
        message(
            '{}/{} primers bind the foreground genome >= {} times'
            .format(results.count(), self.n, min_bind))
        return results

    @_filter
    def filter_max_bg_rate(self, rate):
        """Remove primers that bind more than `rate` to the background genome."""
        results = Primer.select().where(
            (Primer.seq << self.primers) &
            (Primer.bg_freq <= rate))
        message(
            '{}/{} primers bind the background genome <= {} times'
            .format(results.count(), self.n, rate))

        return results

    @_filter
    def filter_tm_range(self, min_tm, max_tm):
        """Remove primers that have melting temps outside this range.

        Finds any missing melt temps for primers.
        """
        self.update_melt_temps()
        results = Primer.select().where(
            (Primer.seq << self.primers) &
            (Primer.tm <= max_tm) &
            (Primer.tm >= min_tm))
        message(
            '{}/{} primers have a melting temp between {} and {} C'
            .format(results.count(), self.n, min_tm, max_tm))
        return results

    @_filter
    def limit_to(self, n):
        """
        Sort by background binding rate, selects the top `n` least frequently
        binding, then returns those ordered by descending bg/fg ratio.
        """
        if n < 1:
            raise ValueError('n must be greater than 1')

        first_pass = (
            Primer.select().where(Primer.seq << self.primers)
            .order_by(Primer.bg_freq)
            .limit(n))

        results = (Primer
                   .select().where(Primer.seq << first_pass)
                   .order_by(Primer.ratio.desc()))

        return results

    @_filter
    def filter_max_gini(self, gini_max, fg_genome_fp):
        """Remove primers with Gini coefficients less than `gini_max`.

        Finds binding locations and Gini coefficients for primers that do not
        have them already.

        :param gini_max: max Gini coefficient (0-1)
        """
        if 0 > gini_max > 1:
            raise ValueError('Gini coefficient must be between 0-1')

        (self
         .update_locations(fg_genome_fp)
         .update_gini(fg_genome_fp))

        results = Primer.select().where(
            (Primer.seq << self.primers) &
            (Primer.gini <= gini_max))

        message(
            '{}/{} primers have a Gini coefficient <= {}'
            .format(results.count(), self.n, gini_max))

        return results

    @_update
    def update_locations(self, fg_genome_fp):
        """Find binding locations for any primers that don't have them."""
        targets = list(Primer.select().where(
            (Primer.seq << self.primers) &
            (Primer._locations >> None)))
        if len(targets) > 0:
            message(
                'Finding binding locations for {} primers...'
                .format(len(targets)))
        for primer in targets:
            primer._update_locations(fg_genome_fp)
        return targets

    @_update
    def update_gini(self, fg_genome_fp):
        """Calculate Gini coef for any primers that don't have it."""
        targets = list(Primer.select().where(
            (Primer.seq << self.primers) &
            (Primer.gini >> None)))
        if len(targets) > 0:
            message(
                'Finding Gini coefficients for {} primers...'
                .format(len(targets)))
        for primer in targets:
            primer._update_gini(fg_genome_fp)
        return targets

    @_update
    def update_melt_temps(self):
        """Calculate melting temp for any primers that don't have it."""
        targets = list(Primer.select().where(
            (Primer.seq << self.primers) &
            (Primer.tm >> None)))
        if len(targets) > 0:
            message(
                'Finding melting temps for {} primers...'
                .format(len(targets)))
        for primer in targets:
            primer.update_tm()
        return targets

    @_update
    def assign_ids(self):
        """Assign sequential ids to active primers.

        Resets any ids previously set.
        """
        Primer.update(_id=-1).execute()
        primers = list(
            Primer.select()
            .where(Primer.seq << self.primers)
            .order_by(Primer.ratio.desc()).execute())
        for i, primer in enumerate(primers):
            primer._id = i + 1
        return primers

    def summarize(self):
        """Output the number of primers currently in list."""
        message('{} primers satisfy all filters so far.'.format(self.n))
        return self

    def activate(self, min_active=1):
        """Activate all the primers in the list.

        :param min_active: The maximum number expected to activate. Warns if
        fewer than this number.
        """
        n = (Primer.update(active=True)
             .where(Primer.seq << self.primers)
             .execute())
        message('Marked {} primers as active.'.format(n))
        if n < min_active:
            message(
                'Note: Fewer than {} primers were selected ({} passed all the '
                'filters). You may want to try less restrictive filtering '
                'parameters.'.format(min_active, n))
        return self

    def _update_n(self):
        n = self.primers.count()
        if n == 0:
            error('No primers left.', exception=False)
        else:
            self.n = n

    @staticmethod
    def _update_in_chunks(targets, chunksize=100, show_progress=True,
                          label='Updating primer db...'):
        """Update the records for the list of primers in chunks.

        Technically, this performs an upsert, where primer records are updated
        or inserted as needed. This is faster than updating the records for
        each primer individually.
        """
        def upsert_chunk(chunk):
            Primer.delete().where(Primer.seq << chunk).execute()
            Primer.insert_many(p.to_dict() for p in chunk).execute()

        chunk_iterator(
            itr=targets,
            fn=upsert_chunk,
            n=chunksize,
            show_progress=show_progress,
            label=label)


def read_primer_list(lines):
    """Read in a list of primers and return their records from the db.

    :param lines: a list of primer sequences, one per line; anything after the
    first whitespace is ignored.
    """
    seqs = [re.split(r'[ \t]+', line.strip('\n'))[0] for line in lines]
    primers = list(Primer.select().where(Primer.seq << seqs).execute())
    if len(primers) < len(seqs):
        primer_seqs = [p.seq for p in primers]
        missing = [_ for _ in seqs if _ not in primer_seqs]
        for seq in missing:
            warn(
                seq + ' not in the database; skipping. Add it manually with '
                '`swga count --input <file>` ')
    return primers



import os
import swga
from swga import quote, __version__
import swga.database
from swga.commands._command import Command
from swga.database import Primer, Set
from swga.utils import fmtkv
import click
from peewee import fn

summary_template = '''
{header}
---------------
  Foreground genome:
    {fg_genome_fp}
  Background genome:
    {bg_genome_fp}
  Excluded sequences:
    {exclude_fp}


PRIMER SUMMARY
---------------
There are {nprimers} primers in the database. {if_no_primers_msg}

{nactive} are marked as active (i.e., they passed filter steps and will be
used to find sets of compatible primers.) {if_no_active_primers_msg}

The average number of foreground genome binding sites is {avg_fg_bind:.0f}.
    (avg binding / genome_length = {fg_bind_ratio:05f})
The average number of background genome binding sites is {avg_bg_bind:.0f}.
    (avg binding / genome_length = {bg_bind_ratio:05f})

{melting_tmp_msg}


SETS SUMMARY
---------------
There are {nsets} sets in the database.
{set_msg}\
'''

best_set_desc = '''\
The best scoring set is #{best_set}, with {bs_size} primers and a score of {bs_score:03f}.
Various statistics:
 {bs_stats}
The primers in Set {best_set} are:
{bs_primers}
'''


class Summary(Command):

    def run(self):
        self.summary_msg = summary_template
        self.best_set_desc = best_set_desc
        avg_fg_bind, avg_bg_bind, nprimers = (
            Primer
            .select(fn.Avg(Primer.fg_freq),
                    fn.Avg(Primer.bg_freq),
                    fn.Count(Primer.seq))
            .scalar(as_tuple=True))

        if (avg_fg_bind is None) or (avg_bg_bind is None):
            (avg_fg_bind, avg_bg_bind) = (0, 0)

        fg_bind_ratio = avg_fg_bind / float(self.fg_length)
        bg_bind_ratio = avg_bg_bind / float(self.bg_length)
        nactive = Primer.select().where(Primer.active == True).count()

        min_tm, max_tm, avg_tm = (
            Primer
            .select(fn.Min(Primer.tm),
                    fn.Max(Primer.tm),
                    fn.Avg(Primer.tm))
            .where(Primer.active == True)
            .scalar(as_tuple=True))

        nsets = Set.select(fn.Count(Set._id)).scalar()

        if nsets > 0:
            bs = Set.select().order_by(Set.score).limit(1).get()
            bs_primers = ", ".join(
                swga.database.get_primers_for_set(bs._id)).strip()
            best_set = bs._id
            bs_size = bs.set_size
            bs_score = bs.score
            bs_stats = "- " + "\n - ".join(
                fmtkv(k, bs.__dict__['_data'][k])
                for k in bs.exported_fields()
                if k not in ["_id", "pids", "score", "primers"]
            )
            self.best_set_desc = self.best_set_desc.format(**locals())

        if_no_primers_msg = click.style(
            "Run `swga count` to find possible primers."
            if nprimers == 0 else "", fg='green')
        if_no_active_primers_msg = click.style(
            "Run `swga filter` to identify primers to use."
            if nactive == 0 else "", fg='green')
        melting_tmp_msg = (
            "The melting temp of the primers ranges between {min_tm:.2f}C and "
            "{max_tm:.2f}C with an average of {avg_tm:.2f}C."
            if nactive > 0 and min_tm and max_tm else
            "No melting temps have been calculated yet.").format(**locals())
        ifzero_sets_msg = click.style(
            "Run `swga find_sets` after identifying valid primers to begin "
            "collecting sets.\n", fg='green')

        set_msg = (self.best_set_desc if nsets > 0 else ifzero_sets_msg)

        primer_db = os.path.abspath(self.primer_db)
        nprimers = click.style(str(nprimers), bold=True, fg='blue')
        nactive = click.style(str(nactive), bold=True, fg='blue')
        nsets = click.style(str(nsets), bold=True, fg='blue')

        self.header = click.style("swga v{}".format(__version__), fg='green')

        # Copy all the relevant values into one dict
        values = self.__dict__.copy()
        values.update(locals())

        # Format the summary message with all the calculated values

        self.summary_msg = self.summary_msg.format(**values)
        click.echo(quote(self.summary_msg, quote="  ", width=200))



# -*- coding: utf-8 -*-
'''
Selects valid primers from a database according to various criteria.

Since all the results of the calculations are stored in the primer db,
the user can run this command multiple times to tune parameters and
subsequent runs will be much faster.
'''

from swga.primers import Primers
from swga.commands._command import Command
import swga.database


class Filter(Command):

    def run(self):
        # If we have an input file, use that. Otherwise pull from db
        if self.input:
            with open(self.input, 'rb') as infile:
                primers = Primers(infile)
        else:
            self.skip_filtering = False
            primers = Primers()

        assert isinstance(primers, Primers)

        # Undo all active marks, if any
        swga.database.Primer.update(active=False).execute()

        if not self.skip_filtering:
            (
                primers
                .filter_min_fg_rate(self.min_fg_bind)
                .filter_max_bg_rate(self.max_bg_bind)
                .summarize()
                .filter_tm_range(self.min_tm, self.max_tm)
                .limit_to(self.max_primers)
                .filter_max_gini(self.max_gini, self.fg_genome_fp)
            )

        primers.activate(self.max_primers)



# -*- coding: utf-8 -*-
"""
`swga init` is a special command that initializes a directory before being used
with SWGA. In short, this reads the foreground and background FASTA files to
get some basic stats (number of records, overall number of bases) and
populates the generated parameters.cfg file with reasonable defaults based on
these stats.

For this reason, this command does not use the framework that other commands
use (since it creates the parameters file).
"""

import click
import subprocess
import argparse
from click._compat import filename_to_ui
import os
import stat
from swga import (
    DEFAULT_CFG_FNAME,
    DEFAULT_DB_FNAME,
    __version__
)
from swga.commands import create_config_file
import swga.database as database
from pyfaidx import Fasta

version = __version__

# Default binding frequencies for foreground and background
# Fg is based on a binding rate of 1/100000 bp/binding site
# Bg is based on a binding rate of 1/150000 bp/binding site
MIN_FG_RATE = 0.00001
MAX_BG_RATE = 0.0000067


setup_msg = '''
swga v{version} - interactive setup
-------------------------------
This will set up a new swga workspace in the current directory ({CWD}).
'''

fg_prompt = '''Enter path to foreground FASTA file'''
fg_parsing = 'Checking {fg_genome_fp}...'
fg_msg = '''\
Foreground filepath: {fg_genome_fp}
  Length:  {fg_length} bp
  Records: {fg_nrecords}
'''

bg_prompt = '''Enter path to background FASTA file'''
bg_parsing = 'Checking {bg_genome_fp}...'
bg_msg = '''\
Background filepath: {bg_genome_fp}
  Length:  {bg_length} bp
'''

excl_prompt = '''\
Do you want to add a FASTA file of sequences that will be used to exclude
primers? For instance, to avoid primers that bind to a mitochondrial genome,
you would add the path to that genome file. There can be multiple sequences in
the file, but only one file can be specified.
'''
excl_prompt2 = '''Enter path to exclusionary sequence(s), in FASTA format'''
excl_msg = '''Exclusionary sequences file: {}'''
excl_msg_no_file = '''No exclusionary sequences file specified.'''

overwrite_cfg_prompt = '''Existing file `{}' will be overwritten. Continue?'''

fin_msg = '''\
Done!

Created pre-filled config file `{}'. Workspace data: `{}'

This file has been pre-filled with reasonable defaults. However, you will
probably want to modify them to suit your needs. You can override the values
in this file both by passing arguments to each command or by modifying this
file in a plain-text editor such as TextEdit or `nano'.
--------------------
'''


def main(argv=None):

    parser = argparse.ArgumentParser(
        description="Initialize the current directory as an swga workspace.")
    parser.add_argument(
        '-f', '--fg_genome_fp',
        help='Path to foreground genome/sequences (in FASTA format)',
        type=argparse.FileType('r'))
    parser.add_argument(
        '-b', '--bg_genome_fp',
        help='Path to background genome/sequences (in FASTA format)',
        type=argparse.FileType('r'))
    parser.add_argument(
        '-e', '--exclude_fp',
        help='Path to sequences to exclude from analysis (in FASTA format)',
        type=argparse.FileType('r'))
    parser.add_argument(
        '--force',
        help='Overwrite any existing data in the directory without prompts',
        action='store_true')

    args = parser.parse_args(argv) if argv else parser.parse_args()

    # Convert to local variables
    fg_genome_fp = args.fg_genome_fp
    bg_genome_fp = args.bg_genome_fp
    exclude_fp = args.exclude_fp

    # Sanity check our default files for read/write
    cwd = os.path.abspath(os.curdir)
    assert os.access(cwd, os.W_OK | os.X_OK)
    db_fp = os.path.abspath(os.path.join(cwd, DEFAULT_DB_FNAME))
    # assert os.access(db_fp, os.W_OK | os.X_OK)
    cfg_fp = os.path.abspath(DEFAULT_CFG_FNAME)
    # assert os.access(cfg_fp, os.W_OK | os.X_OK)

    # 01. Display welcome message
    click.secho(setup_msg.format(version=version, CWD=cwd), fg="blue")

    # 02. Prompt for the foreground genome, if not already specified
    if (not fg_genome_fp):
        fg_genome_fp = click.prompt(
            fg_prompt, type=click.Path(exists=True, resolve_path=True))
    else:
        fg_genome_fp = os.path.abspath(fg_genome_fp.name)
    click.secho(fg_parsing.format(**locals()), fg='blue')
    fg_length, fg_nrecords = fasta_stats(fg_genome_fp)
    click.secho(fg_msg.format(**locals()), fg="green")

    # 03. Prompt for background genome, if not already specified
    if (not bg_genome_fp):
        bg_genome_fp = click.prompt(
            bg_prompt, type=click.Path(exists=True, resolve_path=True))
    else:
        bg_genome_fp = os.path.abspath(bg_genome_fp.name)
    click.secho(bg_parsing.format(**locals()), fg='blue')
    bg_length = fasta_len_quick(bg_genome_fp)
    click.secho(bg_msg.format(**locals()), fg="green")

    # 04. Prompt for a file containing sequences to exclude, if not already
    #     given
    if (not exclude_fp):
        if click.confirm(excl_prompt):
            exclude_fp = click.prompt(
                excl_prompt2, type=click.Path(exists=True, resolve_path=True))
    else:
        exclude_fp = os.path.abspath(exclude_fp.name)

    if exclude_fp:
        click.secho(excl_msg.format(exclude_fp), fg="green")
    else:
        exclude_fp = ""
        click.secho(excl_msg_no_file, fg="green")

    # 05. Build and populate the config file
    default_parameters = create_config_file()
    #cfg_fp = os.path.join(CWD, DEFAULT_CFG_FNAME)
    min_fg_bind = int(MIN_FG_RATE * float(fg_length))
    max_bg_bind = int(MAX_BG_RATE * float(bg_length))

    # 06. Write config file
    if not args.force and os.path.isfile(cfg_fp):
        click.confirm(overwrite_cfg_prompt.format(DEFAULT_CFG_FNAME),
                      abort=True)
    with open(cfg_fp, "wb") as cfg_file:
        cfg_file.write(default_parameters.format(**locals()))

    # 07. Initialize the database
    database.init_db(db_fp, create_if_missing=True)
    database.check_create_tables(db_fp, skip_check=args.force)
    database.set_metadata(
        db_fp,
        version,
        fg_genome_fp,
        bg_genome_fp,
        exclude_fp,
        fg_length,
        bg_length)

    # Done!
    click.secho(fin_msg.format(cfg_fp, db_fp), fg="green")


def fasta_len_quick(fasta_fp):
    """
    Fast way to get the number of bases in a FASTA file, excluding headers.
    """
    try:
        length = subprocess.check_output(
            "grep -v '>' {} | wc -c".format(fasta_fp), shell=True)
        length = int(length.strip())
        return length
    except subprocess.CalledProcessError:
        click.echo(click.style(
            "\nError getting length of FASTA file {}".format(fasta_fp),
            fg="red"))
        raise


def fasta_stats(fasta_fp):
    """
    Retrieves the number of bases and number of records in a FASTA file. Also
    creates a FASTA index (.fai) for later searching. May be slow for very
    large files.
    """
    # pyfaidx can't handle blank lines within records, so we have to check :(
    check_empty_lines(fasta_fp)
    try:
        fasta = Fasta(fasta_fp)
        length = fasta_len_quick(fasta_fp)
        nrecords = len(fasta.keys())
        return length, nrecords
    except:
        click.secho(
            "\nError reading %s: invalid FASTA format?" % fasta_fp, fg="red")
        raise


def check_empty_lines(fasta_fp):
    '''
    The pyfaidx module has issues with blank lines in Fasta files. This checks
    for their presence and offers to remove them in-place.
    '''
    basename = os.path.basename(fasta_fp)
    try:
        check = subprocess.check_output(
            "grep -n '^$' {}".format(fasta_fp), shell=True)
    except subprocess.CalledProcessError:
        check = None
    if check:
        click.confirm(
            "`{}` has blank lines, which can mess up our FASTA indexer."
            " Is it okay to remove these lines from the file?"
            .format(basename), abort=True)
        # sed -ie gets around the fact that BSD sed -i requires a backup extension
        # with sed, but we don't want to create a backup. Ensures linux/osx
        # compatibility.
        subprocess.check_call(
            "sed -ie '/^$/d' \"{}\"".format(fasta_fp), shell=True)

##
# The monkeypatching going on here forces the click framework to strip
# whitespace from the ends of filename strings passed to it. This allows users
# to drag-and-drop files on the Mac OS X terminal (which appends a trailing
# space); without this, click says that the file doesn't exist.
##


def monkeypatch_method(cls):
    def decorator(func):
        setattr(cls, func.__name__, func)
        return func
    return decorator


@monkeypatch_method(click.Path)
def convert(self, value, param, ctx):
    value = value.strip()
    rv = value
    if self.resolve_path:
        rv = os.path.realpath(rv)

    try:
        st = os.stat(rv)
    except OSError:
        if not self.exists:
            return rv
        self.fail('%s "%s" does not exist.' % (
            self.path_type,
            filename_to_ui(value)
        ), param, ctx)

    if not self.file_okay and stat.S_ISREG(st.st_mode):
        self.fail('%s "%s" is a file.' % (
            self.path_type,
            filename_to_ui(value)
        ), param, ctx)
    if not self.dir_okay and stat.S_ISDIR(st.st_mode):
        self.fail('%s "%s" is a directory.' % (
            self.path_type,
            filename_to_ui(value)
        ), param, ctx)
    if self.writable and not os.access(value, os.W_OK):
        self.fail('%s "%s" is not writable.' % (
            self.path_type,
            filename_to_ui(value)
        ), param, ctx)
    if self.readable and not os.access(value, os.R_OK):
        self.fail('%s "%s" is not readable.' % (
            self.path_type,
            filename_to_ui(value)
        ), param, ctx)

    return rv



import functools
import subprocess
import os
import time
import signal

import click
import swga.database as database
import swga.graph as graph
import swga.locate as locate
import swga.score as score
import swga.utils as utils
from swga import (warn, message)
from swga.commands._command import Command
from swga.database import Set


GRAPH_FP = "compatibility_graph.dimacs"
STATUS_LINE = '''\
\rSets: {: ^5,.6g} | Passed: {: ^5,.6g} | Smallest max bind dist:{: ^12,.4G}\
'''


class FindSets(Command):

    def run(self):
        if self.max_sets < 1:
            self.max_sets = float("inf")
        # self.score_cmd = Score(argv)
        # We need to clear all the previously-used sets each time due to
        # uniqueness constraints
        all_sets = Set.select()
        if all_sets.count() > 0:
            if not self.force:
                click.confirm("Remove all previously-found sets?", abort=True)
            for s in all_sets:
                s.primers.clear()
                s.delete_instance()

        self.chr_ends = locate.chromosome_ends(self.fg_genome_fp)
        # Evaluate the scoring expression from a string and return it as a
        # callable function
        self.score_fun = functools.partial(
            score.default_score_set,
            expression=self.score_expression)

        graph.build_graph(self.max_dimer_bp, GRAPH_FP)

        message(
            "Finding sets. If nothing appears, try relaxing your parameters.")

        setfinder_lines = self.find_sets(GRAPH_FP, self.workers)
        self.process_lines(setfinder_lines)

    def find_sets(self, graph_fp, workers):
        if workers <= 1:
            return self._find_sets(graph_fp)
        else:
            return self._mp_find_sets(graph_fp, workers)

    def _find_sets(self, graph_fp, vertex_ordering="weighted-coloring"):
        assert vertex_ordering in ["weighted-coloring", "random"]
        find_set_cmd = [
            utils.set_finder, '-q', '-q',
            '--bg_freq', self.min_bg_bind_dist,
            '--bg_len', self.bg_length,
            '--min', self.min_size,
            '--max', self.max_size,
            '--unweighted',
            '--all',
            '--reorder', vertex_ordering,
            graph_fp
        ]
        find_set_cmd = " ".join([str(_) for _ in find_set_cmd])

        # We call the set_finder command as a line-buffered subprocess that
        # passes its output back to this process.
        # The function then yields each line as a generator; when close() is
        # called, it terminates the set_finder subprocess.
        process = subprocess.Popen(
            find_set_cmd,
            shell=True,
            stdout=subprocess.PIPE,
            preexec_fn=os.setsid,
            bufsize=1)
        try:
            for line in iter(process.stdout.readline, b''):
                (yield line)
        finally:
            time.sleep(0.1)
            if process.poll() is None:
                os.killpg(process.pid, signal.SIGKILL)

    def _mp_find_sets(self, graph_fp, workers):
        setfinder_procs = [self._find_sets(graph_fp, vertex_ordering="random")
                           for _ in range(workers)]
        try:
            while True:
                for i, setfinder in enumerate(setfinder_procs):
                    (yield setfinder.next())
        finally:
            for i, setfinder in enumerate(setfinder_procs):
                setfinder.close()

    def process_lines(self, setfinder_lines):
        passed = processed = 0
        smallest_max_dist = float('inf')

        try:
            for line in setfinder_lines:
                try:
                    primer_ids, bg_dist_mean = score.read_set_finder_line(line)
                except ValueError:
                    warn("Could not parse line:\n\t" + line)
                    continue

                primers = database.get_primers_for_ids(primer_ids)
                processed += 1

                set_score, variables, max_dist = score.score_set(
                    primers=primers,
                    max_fg_bind_dist=self.max_fg_bind_dist,
                    bg_dist_mean=bg_dist_mean,
                    chr_ends=self.chr_ends,
                    score_fun=self.score_fun,
                    interactive=False
                )

                if max_dist < smallest_max_dist:
                    smallest_max_dist = max_dist

                message(
                    STATUS_LINE.format(processed, passed, smallest_max_dist),
                    newline=False)

                # Return early if the set doesn't pass
                if set_score is False:
                    continue
                else:
                    passed += 1

                database.add_set(
                    _id=passed,
                    primers=primers,
                    score=set_score,
                    scoring_fn=self.score_expression,
                    **variables)

                if passed >= self.max_sets:
                    message("\nDone (scored %i sets)" % passed)
                    break
        finally:
            # Raises a GeneratorExit inside the find_sets command, prompting it
            # to quit the subprocess
            setfinder_lines.close()



# -*- coding: utf-8 -*-
'''
Exports primers or sets from the database in a variety of formats.

Examples:

Get set metadata and primers for top 10 sets by score
$ swga export sets --order_by score --descending --limit 10

Get metadata for these four sets
$ swga export sets --ids 1 4 6 12

Get 50 primers with the lowest bg binding frequency
$ swga export primers --order_by bg_freq --limit 50

Get these primers' metadata
$ swga export primers --seq ATGGGCCA GCCATT

Get a set as a BED file for IGV
$ swga export bed_file --set_id 5

Get the Lorenz curve for a set (a visualization of the Gini index)
$ swga export lorenz --set_id 5
'''
import collections
import csv
import os

import swga.database as database
import swga.locate
import swga.stats
from swga.commands._command import Command
from swga.database import Primer, Set
from swga.export import BedGraph, BedFile


class Export(Command):

    def run(self):
        self.header = not self.no_header
        if self.limit < 0:
            self.limit = None
        if self.what in ['set', 'sets']:
            sets = self.get_items(Set)
            export(Set, sets, self.output, self.header)

        if self.what in ['primer', 'primers']:
            primers = self.get_items(Primer)
            export(Primer, primers, self.output, self.header)

        if self.what == 'lorenz':
            sets = self.get_items(Set)
            export_lorenz(sets, self.output, self.fg_genome_fp, self.header)

        if "bed" in self.what:
            outpath = self.output_folder if self.output_folder else os.getcwd()
            sets = self.get_items(Set)
            if self.what == "bedfile":
                for set in sets:
                    swga.message(
                        "Exporting set {} and its primers as bedfiles..."
                        .format(set._id))
                    bedfile = BedFile(set, self.fg_genome_fp)
                    bedfile.write(outpath)
            elif self.what == "bedgraph":
                for set in sets:
                    swga.message("Exporting set {} as bedgraph..."
                                 .format(set._id))
                    bedgraph = BedGraph(
                        set=set,
                        fg_genome_fp=self.fg_genome_fp,
                        opts_str=self.opts_str,
                        window_size=self.window_size,
                        step_size=self.step_size)
                    bedgraph.write(outpath)

    def get_items(self, model):
        '''
        Retrieves the rows from the model given by the arguments (basically an
        adaptor around very simple SQL queries.)
        '''
        ids = self.ids
        order_by = self.order_by
        limit = self.limit
        descending = self.descending

        if ids and (order_by or limit):
            swga.warn("ID(s) specified: ignoring --order_by/--limit options.")
            order_by = limit = None

        validate_order_field(order_by, model)

        if ids:
            targets = model.select().where(model._id << ids)
            for target in targets:
                yield target
        else:
            query = model.select()
            if order_by:
                if descending:
                    query = query.order_by(model.fields()[order_by].desc())
                else:
                    query = query.order_by(model.fields()[order_by])
            if limit:
                query = query.limit(limit)
            targets = query
            for target in targets:
                yield target


def export(model, rows, outfile, header=True):
    '''
    Writes rows from a model (i.e. table) in a tab-delimited format to a file
    specified by outfile. Relies on the custom `exported_fields` function of the
    Primer and Set models.
    '''
    exported_fields = model.exported_fields()

    writer = csv.DictWriter(
        outfile,
        exported_fields,
        delimiter="\t",
        extrasaction='ignore',
        lineterminator='\n'
    )

    if header:
        writer.writeheader()

    for row in rows:
        d = {}
        for field_name in row.exported_fields():
            field = getattr(row, field_name)
            if (not isinstance(field, basestring) and
                    isinstance(field, collections.Iterable)):
                try:
                    field = ",".join([p.seq for p in field])
                except AttributeError:
                    field = ",".join([s._id for s in field])
            d[field_name] = field
        writer.writerow(d)


def export_lorenz(sets, outfile, fg_genome_fp, header=True):
    '''Exports the empirical Lorenz curve used for the given sets.'''
    writer = csv.DictWriter(
        outfile,
        fieldnames=["SetID", "CDF"],
        delimiter="\t",
        lineterminator='\n'
    )

    if header:
        writer.writeheader()

    for set in sets:
        # Get the distances between each primer binding site
        primer_seqs = database.get_primers_for_set(set._id)
        primers = list(Primer.select().where(Primer.seq << primer_seqs).execute())
        chr_ends = swga.locate.chromosome_ends(fg_genome_fp)
        binding_sites = swga.locate.linearize_binding_sites(primers, chr_ends)
        distances = swga.stats.seq_diff(binding_sites)

        lorenz = swga.stats.lorenz(distances)
        lorenz_str = ",".join(str(d) for d in lorenz)
        writer.writerow({'SetID': set._id, 'CDF': lorenz_str})


def validate_order_field(field, model):
    '''Ensures the given field exists in the model.'''
    if field and field not in model.fields():
        swga.error(
            "Cannot order by '{}'. Valid choices are {}"
            .format(field, ", ".join(Primer.fields())))



from ConfigParser import SafeConfigParser
import abc
import click
import argutils
from swga import (
    __version__,
    warn,
    message
)
import swga.utils as utils


class Command(object):

    def __init__(self, name, cfg_file, metadata):
        """Create a new command with the given name and argument string.

        :param name: the command name, used to find the appropriate section in
        the config file
        :param cfg_file: the path to the config file
        :param metadata: a swga.metadata namedtuple generated from
        database.get_metadata()
        """
        spec = utils.specfile(name)
        opts = argutils.read.from_yaml(spec)

        self.name = name
        self.parser = argutils.export.to_argparser(name, opts)
        config = SafeConfigParser()
        if config.read(cfg_file) and config.has_section(name):
            self.parser = argutils.set_parser_defaults(self.parser, config)
        self._set_metadata(metadata)

    @abc.abstractmethod
    def run(self):
        return

    def _set_metadata(self, meta):
        self.primer_db = meta.db_name
        self.fg_genome_fp = meta.fg_file
        self.bg_genome_fp = meta.bg_file
        self.exclude_fp = meta.ex_file
        self.fg_length = meta.fg_length
        self.bg_length = meta.bg_length

    def parse_args(self, argv, quiet=False):
        args, unknown = self.parser.parse_known_args(argv)
        self.args = vars(args)
        self.unknown = unknown
        for k, v in self.args.items():
            setattr(self, k, v)
        if not quiet:
            self._pprint_args()
        if len(self.args) > 0 and all(v is None for k, v in self.args.items()):
            warn(
                "[swga %s]: All parameters are missing- ",
                "this may indicate a corrupt or missing parameters file."
                % self.name)

    def _pprint_args(self):
        message(
            click.style("swga {}, v{}".format(self.name, __version__),
                        fg='green'))
        for arg, val in self.args.iteritems():
            if val is None or val == "":
                val = click.style("None", fg='red')
            message(click.style("  - {}: {}".format(arg, val), fg='blue'))



import os
from types import *
from collections import Counter
from swga import (warn, message)
import swga.database as database
import swga.locate as locate


def _mk_folder(folder, fg_genome_fp, setstr):
    fg_name = ".".join(
        os.path.basename(fg_genome_fp).split(".")[:-1]) + "_export"
    output_folder = os.path.join(folder, fg_name, setstr)
    if not os.path.isdir(output_folder):
            os.makedirs(output_folder)
    return output_folder


class BedFile(object):
    """Creates and writes BedFiles from a set of primers."""
    def __init__(self, set, fg_genome_fp):
        super(BedFile, self).__init__()
        assert isinstance(set, database.Set)

        self.set = set
        self.fg_genome_fp = fg_genome_fp

    def _primer_sites(self, primer):
        seq = primer.seq
        for record_name, locations in primer.locations.iteritems():
            for location in locations:
                yield record_name, location, location + len(seq)

    def write(self, output_fp):
        '''
        Writes n+1 BED-formatted files to a folder, where n is the number of
        primers in a set. One file is written for each primer, and then one
        file is written that combines all the primers together into one set.
        '''
        # Create the output folder if it doesn't exist already
        output_folder = _mk_folder(
            output_fp,
            self.fg_genome_fp,
            "set_%s" % self.set._id)

        whole_set_fp = os.path.join(output_folder, 'whole_set.bed')

        with open(whole_set_fp, 'wb') as whole_set_file:
            whole_set_file.write("track name=Set_{}\n".format(self.set._id))
            for primer in self.set.primers:
                seq = primer.seq
                primer_fp = os.path.join(output_folder, seq + ".bed")
                with open(primer_fp, 'wb') as primer_file:
                    primer_file.write("track name={}\n".format(seq))
                    for record_name, start, end in self._primer_sites(primer):
                        recordstr = "{} {} {}\n".format(
                            record_name, start, end)
                        primer_file.write(recordstr)
                        whole_set_file.write(recordstr)
        message("Bedfiles written to "+output_folder)


class BedGraph(object):
    """
    Creates a BedGraph-format file from a given set that illustrates the
    amount of coverage the primers in that set give over the foreground
    genome.
    """

    def __init__(self, set, fg_genome_fp, opts_str, window_size, step_size):
        '''
        args:
        set: the set of primers to export
        fg_genome_fp: the path to the foreground genome
        opts_str: the string to write after `track type=bedGraph`
        window_size: the size of the sliding window (<= size of each chromosome or record)
        step_size: the step size of the sliding window (must be smaller than window_size)
        '''
        super(BedGraph, self).__init__()
        assert isinstance(set, database.Set)

        self.set = set
        self.fg_genome_fp = fg_genome_fp
        self.opts_str = opts_str
        self.window_size = int(window_size)
        self.step_size = int(step_size) if step_size else self.window_size / 5

    def _hits_per_record(self):
        '''
        Calculates the number of bases covered by primers in each record, and
        yields the record name, midpoint of the record, and number of hits as
        a tuple.
        '''
        record_ends = locate.chromosome_ends(self.fg_genome_fp)
        for record_name, ends in record_ends.iteritems():
            record_length = ends[1] + 1

            # Check window size <= record_length and fix if not
            this_window_size = self.window_size
            if this_window_size > record_length:
                this_window_size = record_length
                warn(
                    "In [{}]: window size larger than record; set to {}"
                    .format(record_name, this_window_size))

            # Check step size is compatible with window size and fix if not
            this_step_size = self.step_size
            if this_step_size > this_window_size:
                this_step_size = this_window_size
                warn(
                    "In [{}]: step size larger than window size ({}), set to {}"
                    .format(record_name, this_window_size, this_step_size))

            # Count the number of primers that bind in the current record
            counter = Counter()
            for primer in self.set.primers:
                counter.update(Counter(primer.locations[record_name]))

            starting_positions = xrange(
                0, int(record_length - this_window_size), this_step_size)

            for start in starting_positions:
                end = start + this_window_size
                midpoint = (end + start) / 2
                # Add each base's count to get the number of bases covered
                hits = sum([counter[i] for i in xrange(start, end)])

                yield record_name, midpoint, hits

    def write(self, output_fp):
        """Writes the bedgraph to a file in a directory named after the set."""
        # Create the output folder if it doesn't exist already
        output_folder = _mk_folder(
            output_fp,
            self.fg_genome_fp,
            "set_%s" % self.set._id)

        bedgraph_fp = os.path.join(
            output_folder,
            "set_{}.bedgraph".format(self.set._id))

        with open(bedgraph_fp, 'wb') as bedgraph_file:
            typestr = "track type=bedGraph {}\n".format(self.opts_str)
            bedgraph_file.write(typestr)
            for record_name, midpoint, hits in self._hits_per_record():
                linestr = "{} {} {} {}\n".format(
                    record_name, midpoint, midpoint, hits)
                bedgraph_file.write(linestr)

        message("Bedgraph written to {}".format(bedgraph_fp))


from __future__ import absolute_import
from collections import namedtuple
import pkg_resources
import textwrap
import sys

import click


DEFAULT_DB_FNAME = "swga.db"
DEFAULT_CFG_FNAME = "parameters.cfg"

__version__ = pkg_resources.require("swga")[0].version


# A struct-like object that holds basic information about the workspace
meta = namedtuple(
    "meta",
    "db_name fg_file bg_file ex_file fg_length bg_length")


def error(msg, exception=True, wrap=True):
    '''Prints an error message to stderr and exits.'''
    if exception:
        raise SWGAError(msg)
    else:
        msg = quote(msg, width=75) if wrap else msg
        click.secho(msg, fg='red', err=True)
        sys.exit(1)


def warn(msg):
    '''Prints a warning message to stderr.'''
    text = quote(msg, quote="!> ", nl=False)
    click.secho(text, err=True, fg='red')


def message(msg, newline=True):
    click.echo(msg, err=True, nl=newline)


def quote(text, width=72, quote="", nl=True):
    if not text:
        return ""
    out = ""
    for line in text.split("\n"):
        sublines = textwrap.wrap(line, width=width, replace_whitespace=False)
        sublines = [quote + l for l in sublines]# if l.strip()]
        out += "\n".join(sublines) + "\n"
    return out


class SWGAError(Exception):
    pass



import os
import errno
import swga
from click import progressbar
from pkg_resources import (
    resource_exists, resource_filename, resource_stream
)

__all__ = [
    "mkdirp",
    "chunk_iterator",
    "specfile",
    "dsk",
    "set_finder"
]


def mkdirp(path):
    """Create a directory unless it already exists, using EAFP methods."""
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def chunks(l, n):
    for i in xrange(0, len(l), n):
        yield l[i:i + n]


def chunk_iterator(itr, fn, n=100, show_progress=True, label=None):
    """Break an iterable into chunks and apply a function to each chunk.

    :param itr: the iterable to be chunked
    :param fn: the function to be applied to each chunks
    :param n: the size of each chunk
    :param show_progress: show a progress bar
    :param label: the label to show on the progress bar
    """
    length = len(itr)
    if length == 0:
        return
    label = "" if label is None else label
    if length / n <= 1:
        show_progress = False
        swga.message(label)
    chunked_itr = chunks(itr, n)
    if show_progress:
        with progressbar(chunked_itr, max(length / n, 1), label) as bar:
            for chunk in bar:
                fn(chunk)
    else:
        for chunk in chunked_itr:
            fn(chunk)


def _get_resource_file(rs):
    _rs = os.path.join('bin', rs)
    if resource_exists("swga", _rs):
        res_path = resource_filename("swga", _rs)
        return res_path
    else:
        swga.error("Could not find `{}': try re-installing swga.".format(rs))


def specfile(name):
    """Return the path to the specfile for a given command name."""
    fp = os.path.join('commands', 'specfiles', name + '.yaml')
    return resource_stream("swga", fp)


def fmtkv(k, v):
    """Pretty-print a key-value pair."""
    set_stat_line = "{key:.<16}: {val: >10s}"
    num_fmt_str = "{:,G}"
    if not isinstance(v, basestring):
        print k, v
        v = num_fmt_str.format(v)
    return set_stat_line.format(key=k, val=v)


dsk = _get_resource_file('dsk')
set_finder = _get_resource_file('set_finder')



import os
import re
import platform
import subprocess
from setuptools import setup, find_packages, Command, dist
from distutils.command.build import build as _build
from distutils.command.build_py import build_py as _build_py
from distutils.spawn import find_executable
from setuptools.command.bdist_egg import bdist_egg as _bdist_egg

version = "0.4.0"

class BinaryDistribution(dist.Distribution):
    def is_pure(self):
        return False



class bdist_egg(_bdist_egg):
    def run(self):
        self.run_command('build_py')
        _bdist_egg.run(self)


class build_py(_build_py):

    def run(self):
        """
        Builds `DSK` and `cliquer` in the contrib/ directory and moves them 
        to the right place before installing `swga`.
        """
        def find_prog(target='g++'):
            '''Finds all items on the path with the target in their name.'''
            for path in os.environ["PATH"].split(os.pathsep):
                path = path.strip("")
                if not os.path.isdir(path):
                    continue
                for item in os.listdir(path):
                    if target in item:
                        yield item

        def get_version(target, string):
            '''Finds a version number (i.e. gcc-4.9 ==> 4.9)'''
            match = re.match(r'{}-([0-9.]+)'.format(re.escape(target)), string)
            if match:
                return float(match.group(1))

        # Defaults (e.g. for Linux)
        CC = find_executable('g++')
        omp = 1
        osx = 0
        # Check if we're on a Mac and adjust compiler accordingly
        if 'Darwin' in platform.system() and not os.environ.get("CC"):
            osx = 1
            gcc_version = subprocess.check_output(
                [CC, '--version'], 
                stderr=subprocess.STDOUT
            )
            if 'LLVM' in gcc_version:
                # We'd rather not use clang, so search the path for the most 
                # recent version of g++ we can find
                bestv = (0, '')
                for hit in find_prog("g++"):
                    ver = get_version("g++", hit)
                    if ver >= bestv[0]:
                        bestv = (ver, hit)
                real_CC = find_executable(bestv[1])
                if real_CC:
                    print(
                        "\nNote: Found real g++ living in `{}`, using it instead "
                        "of clang. Override by setting CC=/path/to/gcc before "
                        "this command.\n".format(real_CC)
                    )
                    CC = real_CC
                else:
                    print(
                        "\nFound clang but no GNU g++ compiler: Disabling "
                        "multiprocessing for find_sets. If GNU g++ is "
                        "installed, set it with CC=/path/to/gcc before this "
                        "command."
                    )
                    omp=0

        cliquer_make_cmd = (['make'])
        dsk_make_cmd = (
            'make',
            'CC='+CC,
            'omp='+str(omp),
            'osx='+str(osx)
        )

        print(" ".join(cliquer_make_cmd))
        if not self.dry_run:
            subprocess.check_call(cliquer_make_cmd, cwd="contrib/cliquer")
        print(" ".join(dsk_make_cmd))
        if not self.dry_run:
            subprocess.check_call(dsk_make_cmd, cwd="contrib/dsk")

        if not self.dry_run:            
            try:
                subprocess.check_call(["mkdir", "-p", "swga/bin"])
                print("Moving set_finder")
                os.rename("contrib/cliquer/set_finder", "swga/bin/set_finder")
                print("Moving dsk")
                os.rename("contrib/dsk/dsk", "swga/bin/dsk")
            except OSError:
                raise

        _build_py.run(self)

setup(
    name='swga',
    version=version,
    author='Erik Clarke',
    author_email='ecl@mail.med.upenn.edu',
    packages=find_packages(),
    install_requires=[
        'pyfaidx',
        'click',
        'pyyaml',
        'peewee',
        'melt',
        'pytest',
        'argutils',
        'semantic_version'
    ],
    classifiers = [
        'Programming Language :: Python',
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License (GPL)'
    ],
    package_data={
        'swga': ['commands/specfiles/*', 'bin/set_finder', 'bin/dsk']
    },
    data_files=[ 
        ('bin', ['swga/bin/set_finder', 'swga/bin/dsk'])
    ],
    eager_resources=["bin/set_finder", "bin/dsk"],
    include_package_data=True,
    distclass=BinaryDistribution,
    url='https://github.com/eclarke/swga',
    license='LICENSE.txt',
    description='Pipeline to select compatible primer sets for selective whole-genome amplification.',
    long_description=open('README.md').read(),
    entry_points = {'console_scripts': ['swga = swga.commands.swga_entry:main']},
    cmdclass={
        'bdist_egg': bdist_egg,
        'build_py': build_py
    }
)



import itertools
from swga.kmers import max_sequential_nt
from swga import (error, message)
from swga.primers import Primers


def build_edges(starting_primers, max_binding):
    '''
    Adds a primer pair to the list of edges if it passes the heterodimer
    filter using the max_binding cutoff.
    '''
    edges = []
    for p1, p2 in itertools.combinations(starting_primers, 2):
        if (p1.seq not in p2.seq) and (p2.seq not in p1.seq):
            if max_sequential_nt(p1.seq, p2.seq) <= max_binding:
                edges.append([p1._id, p2._id])
    return edges


def write_graph(primers, edges, file_handle):
    '''
    Writes graph in DIMACS graph format, specified in the cliquer user manual.
    See http://users.tkk.fi/~pat/cliquer.html

    An edge is a list of the form [first_node, second_node]
    "edges" is a list of edges
    '''

    if type(file_handle) is not file:
        raise ValueError(("file_handle must be a file, not "
                          "a {}").format(type(file_handle)))

    file_handle.write('p sp {} {}\n'.format(len(primers), len(edges)))
    for primer in primers:
        weight = primer.bg_freq if primer.bg_freq > 0 else 1
        file_handle.write('n {} {}\n'.format(primer._id, weight))
    for edge in edges:
        try:
            file_handle.write('e {} {}\n'.format(edge[0], edge[1]))
        except IndexError:
            raise ValueError("Edges must be specified as a list with" +
                             "two elements. Invalid edge: {}".format(edge))


def build_graph(max_hetdimer_bind, outfile):
    '''Selects all active primers and outputs a primer compatibility graph.'''

    # Reset all the primer IDs (as ids are only used for set_finder)
    primers = Primers.select_active().assign_ids()
#    print [(p._id, p.ratio) for p in primers]
    message("Composing primer compatibility graph...")
    edges = build_edges(primers, max_hetdimer_bind)

    if len(edges) == 0:
        error(
            "No compatible primers. Try relaxing your parameters.",
            exception=False)

    with open(outfile, 'wb') as out:
        write_graph(primers, edges, out)
